""" Render pipeline """

import logging
import pydot
from .models import KEGGPathway
from .resolver import KEGGPathwayResolver
from .utils import ColorGradient
from typing import Any, Dict, Optional, List

class KEGGPathwayRenderer:
    """
    Renderer for KEGG Pathway
    """
    def __init__(self, kegg_pathway: KEGGPathway):
        self.pathway = kegg_pathway
        self.overlay: Dict[int, Any] = {}
        self.exp_min = 0
        self.exp_max = 0
        self.render_string = None

        # TODO: fix any typing
        self.cmap_upreg: List[Any] = []
        self.cmap_downreg: List[Any] = []

        self.upper_color = (255, 0, 0)
        self.lower_color = (0, 0, 255)

        self.components = KEGGPathwayResolver.get_components()

    def set_overlay(self, gene_dict: dict):
        """
        Overlay dot rendering with differential expression data. {<gene-id>: <fold-exp>}
        :param gene_dict: dict
        :return:
        """
        self.cmap_downreg = ColorGradient(start=(255, 255, 255), stop=self.lower_color, steps=100).get_list()
        self.cmap_upreg = ColorGradient(start=(255, 255, 255), stop=self.upper_color, steps=100).get_list()
        self.overlay = gene_dict
        self.exp_min = min(gene_dict.values())
        self.exp_max = max(gene_dict.values())

    def _get_gene_color(self, gene_id: int):
        if gene_id not in self.overlay:
            return "#ffffff"
        else:
            if self.overlay[gene_id] < 0:
                return self.cmap_downreg[abs(int(self.overlay[gene_id] / self.exp_min * 100))]
            else:
                return self.cmap_upreg[abs(int(self.overlay[gene_id] / self.exp_max * 100))]

    def group_render(self):
        stringBuilder = []
        stringBuilder.append("digraph pathway{CODE} {{".format(CODE=self.pathway.number))
        stringBuilder.append("\tnode [shape=rectangle arrowhead=normal];")
        stringBuilder.append("\tlabel=\"{TITLE}\";".format(TITLE=self.pathway.title))
        stringBuilder.append("\tfontsize=25;\n\tlabelloc=\"t\";")

        # search for "path:<...>"
        # TODO : divide entries in relation groups [[<entry>, ...], [...]]

        entry_groups = []
        genes_of_interest = []

        for rel in self.pathway.relations:
            if len(entry_groups) == 0:
                entry_groups.append([int(rel.entry1), int(rel.entry2)])
            else:
                conn = []
                for n in range(0, len(entry_groups)):
                    if int(rel.entry1) in entry_groups[n] or int(rel.entry2) in entry_groups[n]:
                        conn.append(n)
                if len(conn) == 0:
                    entry_groups.append([int(rel.entry1), int(rel.entry2)])
                elif len(conn) == 1:
                    entry_groups[conn[0]].extend([int(rel.entry1), int(rel.entry2)])
                else:
                    # more then 1 match -> merge groups
                    entry_groups[conn[0]].extend(entry_groups[conn[1]])
                    del entry_groups[conn[1]]
                    entry_groups[conn[0]].extend([int(rel.entry1), int(rel.entry2)])
        # print(entry_groups)

        # TODO : entry group in cluster (subgraph) --> annotate cluster with pathway

        for n in range(0, len(entry_groups)):
            stringBuilder.append("\tsubgraph cluster{N} {{".format(N=n))
            stringBuilder.append("\tlabel=\"\";\n\tcolor=blue;\n\tstyle=dashed;")

            for group_item in entry_groups[n]:
                for entry in self.pathway.entries:
                    # TODO : shape=oval for components --> get comp name
                    if int(entry.id) == group_item:
                        """stringBuilder.append("\tnode [label=\"{LABEL}\"]; entry{ID};"
                                             .format(LABEL=entry.graphics.name.split(", ")[0], ID=entry.id))"""
                        if entry.type == "gene":
                            # shape=rectangle, style=filled, fillcolor=\"#cccccc\"
                            stringBuilder.append("\tnode [label=\"{LABEL}\", shape=rectangle,"
                                                 " style=filled, fillcolor=\"#ffffff\"]; entry{ID};"
                                                 .format(LABEL=entry.graphics.name.split(", ")[0], ID=entry.id))
                        elif entry.type == "group":
                            # shape=rectangle, color="black" label=<<table border='0' cellborder='1'>
                            # <tr><td>comp 1</td></tr><tr><td>comp 2</td></tr></table>>];
                            labels = [self.pathway.get_entry_by_id(comp.id).graphics.name.split(", ")[0] for comp in
                                      entry.components]
                            s_label = ["<tr><td>{LABEL}</td></tr>".format(LABEL=l) for l in labels]
                            stringBuilder.append("\tnode [label=<<table border='0' cellborder='1'>{LABELS}</table>>,"
                                                 " shape=rectangle,"
                                                 " style=filled, color=black, fillcolor=\"#ffffff\"]; entry{ID};"
                                                 .format(LABELS="".join(s_label), ID=entry.id))
                        elif entry.type == "compound":
                            stringBuilder.append("\tnode [label=\"{LABEL}\", shape=oval,"
                                                 " style=filled, fillcolor=\"#ffffff\"]; entry{ID};"
                                                 .format(LABEL=entry.graphics.name.split(", ")[0], ID=entry.id))
                        genes_of_interest.append(entry.id)

            stringBuilder.append("\t}")

        for rel in self.pathway.relations:
            # TODO : adjust arrowhead
            if rel.entry1 in genes_of_interest and rel.entry2 in genes_of_interest:
                stringBuilder.append("\tentry{ID1} -> entry{ID2};".format(ID1=rel.entry1, ID2=rel.entry2))

        stringBuilder.append("}")
        self.render_string = "\n".join(stringBuilder)
        return self.render_string

    def render(self):
        """
        digraph G {
            labelloc="t";
            label="Inflammatory bowel disease (IBD)";
            fontsize=25;
            node [shape=rectangle arrowhead=normal];
            node [label="Tlr2"]; entry1;
            node [label="Nfkb1"]; entry2;
            node [label="Nod2"]; entry3;
            entry1 -> entry2;
            entry1 -> entry3 [arrowhead="tee"];
        }
        """
        stringBuilder = []
        stringBuilder.append("digraph pathway{CODE} {{".format(CODE=self.pathway.number))
        stringBuilder.append("\tnode [shape=rectangle arrowhead=normal];")
        stringBuilder.append("\tlabel=\"{TITLE}\";".format(TITLE=self.pathway.title))
        stringBuilder.append("\tfontsize=25;\n\tlabelloc=\"t\";")

        # TODO : specifiy for each node: shape=circle|rectangle|oval|ellipse, style=filled, fillcolor=red|green|#cccccc

        # search for "path:<...>"

        # TODO : parse entry::type[group]

        related_entries = [int(p.entry1) for p in self.pathway.relations]
        related_entries.extend([int(p.entry2) for p in self.pathway.relations])

        for entry in self.pathway.entries:
            # only render genes with at least 1 relation
            if int(entry.id) in related_entries:  # and entry.type == "gene"
                # case select for types gene, comp, group, ...
                if entry.type == "gene":
                    # shape=rectangle, style=filled, fillcolor=\"#cccccc\"
                    stringBuilder.append("\tnode [label=\"{LABEL}\", shape=rectangle,"
                                         " style=filled, fillcolor=\"#ffffff\"]; entry{ID};"
                                         .format(LABEL=entry.graphics.name.split(", ")[0], ID=entry.id))
                elif entry.type == "group":
                    # shape=rectangle, color="black" label=<<table border='0' cellborder='1'>
                    # <tr><td>comp 1</td></tr><tr><td>comp 2</td></tr></table>>];
                    labels = [self.pathway.get_entry_by_id(comp.id).graphics.name.split(", ")[0] for comp in entry.components]
                    s_label = ["<tr><td>{LABEL}</td></tr>".format(LABEL=l) for l in labels]
                    stringBuilder.append("\tnode [label=<<table border='0' cellborder='1'>{LABELS}</table>>,"
                                         " shape=rectangle,"
                                         " style=filled, color=black, fillcolor=\"#ffffff\"]; entry{ID};"
                                         .format(LABELS="".join(s_label), ID=entry.id))
                elif entry.type == "compound":
                    stringBuilder.append("\tnode [label=\"{LABEL}\", shape=oval,"
                                         " style=filled, fillcolor=\"#ffffff\"]; entry{ID};"
                                         .format(LABEL=entry.graphics.name.split(", ")[0], ID=entry.id))

        for rel in self.pathway.relations:
            # TODO : adjust arrowhead
            stringBuilder.append("\tentry{ID1} -> entry{ID2};".format(ID1=rel.entry1, ID2=rel.entry2))
        stringBuilder.append("}")

        self.render_string = "\n".join(stringBuilder)

        return self.render_string

    def raw_render(self):
        stringBuilder = []
        stringBuilder.append("digraph pathway{CODE} {{".format(CODE=self.pathway.number))

        stringBuilder.append("\tgraph [fontname = \"arial\"];\n\tnode [fontname = \"arial\"];")
        stringBuilder.append("\tedge [fontname = \"arial\"];")

        stringBuilder.append("\tnode [shape=rectangle arrowhead=normal];")
        stringBuilder.append("\tlabel=\"{TITLE}\";".format(TITLE=self.pathway.title))
        stringBuilder.append("\tfontsize=25;\n\tlabelloc=\"t\";")

        related_entries = [int(p.entry1) for p in self.pathway.relations]
        related_entries.extend([int(p.entry2) for p in self.pathway.relations])

        for entry in self.pathway.entries:
            # only render genes with at least 1 relation
            if int(entry.id) in related_entries:
                if entry.type == "gene":
                    stringBuilder.append("\tnode [label=\"{LABEL}\", shape=rectangle,"
                                         " style=filled, fillcolor=\"{COLOR}\"]; entry{ID};"
                                         .format(LABEL=entry.graphics.name.split(", ")[0], ID=entry.id,
                                                 COLOR=self._get_gene_color(entry.get_gene_id())))
                elif entry.type == "group":
                    labels = [self.pathway.get_entry_by_id(comp.id).graphics.name.split(", ")[0] for comp in
                              entry.components]
                    s_label = ["<tr><td>{LABEL}</td></tr>".format(LABEL=l) for l in labels]
                    stringBuilder.append("\tnode [label=<<table border='0' cellborder='1'>{LABELS}</table>>,"
                                         " shape=rectangle,"
                                         " style=filled, color=black, fillcolor=\"#ffffff\"]; entry{ID};"
                                         .format(LABELS="".join(s_label), ID=entry.id))
                elif entry.type == "compound":
                    name = entry.graphics.name.split(", ")[0]
                    if name in self.components:
                        name = self.components.get(name)
                    stringBuilder.append("\tnode [label=\"{LABEL}\", shape=oval,"
                                         " style=filled, fillcolor=\"#ffffff\"]; entry{ID};"
                                         .format(LABEL=name, ID=entry.id))

        for rel in self.pathway.relations:
            # adjust arrowhead
            label = ""
            arrowhead = "normal"
            if "inhibition" in rel.subtypes or "repression" in rel.subtypes:
                arrowhead = "tee"
            elif "binding/association" in rel.subtypes:
                arrowhead = "none"

            if "phosphorylation" in rel.subtypes:
                label = "+p"

            stringBuilder.append("\tentry{ID1} -> entry{ID2} [arrowhead=\"{ARROW}\" label=\"{LABEL}\"];".format(ID1=rel.entry1, ID2=rel.entry2, ARROW=arrowhead, LABEL=label))

        stringBuilder.append("}")
        self.render_string = "\n".join(stringBuilder)
        return self.render_string

    def to_dot(self):
        return pydot.graph_from_dot_data(self.render_string)[0]

    def export(self, extension: str):
        # render with pydot to png|svg|pdf|
        if not self.render_string:
            self.render()
        return self.to_dot().create(format=extension)

    def render_legend(self):
        return """<?xml version="1.0" standalone="no"?>
                    <svg height="200" width="300" version="1.1" baseProfile="full" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" xmlns:ev="http://www.w3.org/2001/xml-events">
                    <defs>
                        <linearGradient id="cmap" x1="0%" y1="0%" x2="0%" y2="100%">
                            <stop offset="0%" style="stop-color:{COLOR_UP};stop-opacity:1" />
                            <stop offset="50%" style="stop-color:rgb(255,255,255);stop-opacity:1" />
                            <stop offset="100%" style="stop-color:{COLOR_DOWN};stop-opacity:1" />
                        </linearGradient>
                    </defs>
                    <g>
                        <rect x="20" y="50" width="20" height="100" fill="url(#cmap)" />
                        <rect x="20" y="50" width="20" height="100" style="stroke:black;stroke-width:2;fill-opacity:0;stroke-opacity:1" />
                        <text x="55" y="150" fill="black" alignment-baseline="central">{LOWER}</text>
                        <text x="55" y="100" fill="black" alignment-baseline="central">0</text>
                        <text x="55" y="50" fill="black" alignment-baseline="central">{UPPER}</text>

                        <line x1="40" y1="50" x2="50" y2="50" style="stroke:rgb(0,0,0);stroke-width:2" />
                        <line x1="40" y1="100" x2="50" y2="100" style="stroke:rgb(0,0,0);stroke-width:2" />
                        <line x1="40" y1="150" x2="50" y2="150" style="stroke:rgb(0,0,0);stroke-width:2" />
                    </g>
                    </svg>""".format(LOWER=self.exp_min, UPPER=self.exp_max,
                                     COLOR_UP=ColorGradient.to_css(color=self.upper_color),
                                     COLOR_DOWN=ColorGradient.to_css(color=self.lower_color))


if __name__ == "__main__":
    pathway = KEGGPathwayResolver(org="mmu").get_pathway("04659")

    renderer = KEGGPathwayRenderer(kegg_pathway=pathway)
    # print(renderer.group_render())
    # print(renderer.render())
    print(renderer.raw_render())





""" KEGG Enrichment analysis core """

import logging
import csv
import os
from .models import KEGGPathway
from .storage import KEGGDataStorage
from .resolver import KEGGPathwayResolver
from typing import List, Any, Dict, NoReturn, Optional, Union
import scipy.stats as stats
from io import TextIOWrapper, StringIO


# def get_all_pathways(org: str):
#     filename = "pathways_{ORG}.dump".format(ORG=org)
#     if KEGGDataStorage.exist(filename=filename):
#         return KEGGDataStorage.load_dump(filename=filename)
#     else:
#         resolve = KEGGPathwayResolver(org=org)
#         pathways = resolve.get_pathway_list()
#         KEGGDataStorage.save_dump(filename=filename, data=pathways)
#         return pathways


class KEGGPathwayAnalysisResult:
    def __init__(self, org: str, pathway_id: str, pathway_name: str, found_genes: list, pathway_genes: list):

        self.organism: str = org
        self.pathway_id = pathway_id
        self.pathway_name = pathway_name
        self.found_genes = found_genes
        self.study_count = len(found_genes)
        self.pathway_genes = pathway_genes
        self.pathway_genes_count = len(pathway_genes)
        self.pvalue : Union[float, None] = None

        # self.args = {
        #     "organism": org,
        #     "pathway_id": pathway_id,
        #     "pathway_name": pathway_name,
        #     "found_genes": found_genes,
        #     "study_count": len(found_genes),
        #     "pathway_genes": pathway_genes,
        #     "pathway_genes_count": len(pathway_genes),
        #     "pvalue": "",
        # }

    def __str__(self):
        return f"<KEGGPathwayAnalysisResult {self.organism}:{self.pathway_id} ({pathway_name}) {self.found_genes}/{self.pathway_genes_count}>"


    def __repr__(self):
        print(self.__str__())

    def set_pvalue(self, pval: float):
        # self.args["pvalue"] = pval
        self.pvalue = pval


    def json_summary(self, gene_delimiter=","):

        # result = list()
        # result.append(self.args["pathway_name"])
        # result.append(self.args["pathway_id"])
        # result.append(str(len(self.args["found_genes"])))
        # result.append(str(len(self.args["pathway_genes"])))
        # result.append(str(self.args["pvalue"]))
        # result.append(",".join([str(a) for a in self.args["found_genes"]]))
        # # result.append(",".join([str(a) for a in self.args["pathway_genes"]]))
        # return "\t".join(result)

        result = {
            "pathway_name": self.pathway_name,
            "pathway_id": self.pathway_id,
            "study_count": self.study_count,
            "pathway_genes": self.pathway_genes_count,
            "pvalue": self.pvalue,
            "found_genes": gene_delimiter.join([str(a) for a in self.found_genes])
        }

        return result

    @staticmethod
    def get_header():
        # return list(self.args.keys())
        return ["pathway_name", "pathway_id", "study_count", "pathway_genes", "pvalue", "found_genes"]




class KEGGPathwayAnalysis:
    def __init__(self, org: str, pathways=None):
        self.organism = org
        self.summary: List[Any] = []
        self.resolver = KEGGPathwayResolver(self.organism)

        # self.all_pathways = get_all_pathways(org=self.organism)
        self.all_pathways = {}

        if pathways == None:
            # If pathways are not set use all pathways
            self.all_pathways = self.resolver.get_pathway_list()

        elif type(pathways) == dict:
            # Use given data when dict is passed
            self.all_pathways = pathways

        elif type(pathways) == list:
            # Load all pathways and filter for pathways in list
            for key, value in self.resolver.get_pathway_list().items():
                if key in pathways:
                    self.all_pathways[key] = value
            if len(self.all_pathways.keys()) == 0:
                raise ValueError("Pathways filter does not succeed. Still 0 pathways in list.")

        else:
            # still not pathways, raise error
            raise ValueError("Pass list or dict of pathways to filter list.")


    def _check_analysis_result_exist(self):
        # Check if summary exists
        if not self.summary or len(self.summary) == 0:
            raise ValueError("need to 'run_summary' first")


    def get_subset(self, subset: list, inplace=False):
        """
        Create subset of analysis result by list of pathway ids
        :param subset: list
        :param inplace: bool
        """
        self._check_analysis_result_exist()

        buffer = []
        subset = [str(s) for s in subset]
        for item in self.summary:
            if str(item.pathway_id) in subset:
                buffer.append(item)

        if inplace:
            self.summary = buffer
        else:
            return buffer


    def run_analysis(self, gene_list: list):
        """
        List of gene ids. Return list of KEGGPathwayAnalysisResult instances
        :param gene_list: list
        :return: list(KEGGPathwayAnalysisResult)
        """

        result = []
        all_found_genes = 0
        absolute_pathway_genes = 0
        study_n = len(gene_list)

        for pathway_id, name in self.all_pathways.items():

            pathway = self.resolver.get_pathway(code=pathway_id)
            # pathway = KEGGPathway.parse(KEGGDataStorage.load_pathway("mmu", pathway_id))

            genes_found = []
            all_pathways_genes = pathway.get_genes().keys()
            absolute_pathway_genes += len(all_pathways_genes)

            for gene_id in all_pathways_genes:
                if gene_id in gene_list:
                    genes_found.append(gene_id)

            all_found_genes += len(genes_found)
            pathway_result = KEGGPathwayAnalysisResult(org="mmu", pathway_id=pathway_id, pathway_name=name, found_genes=genes_found, pathway_genes=all_pathways_genes)
            result.append(pathway_result)

        # Perform Fisher exact test
        # http://docs.scipy.org/doc/scipy-0.17.0/reference/generated/scipy.stats.fisher_exact.html
        for analysis in result:

            # Skip p value calculation if no genes are found
            if analysis.study_count > 0:

                a_var = analysis.study_count
                b_var = study_n - analysis.study_count
                c_var = analysis.pathway_genes_count - analysis.study_count
                d_var = absolute_pathway_genes - analysis.pathway_genes_count - b_var

                # print(a_var, b_var, "\n", c_var, d_var, "\n" + str(absolute_pathway_genes) + "-" * 10)

                _, pval = stats.fisher_exact(
                    [
                        [
                            a_var,
                            b_var
                        ],
                        [
                            c_var,
                            d_var
                        ]
                    ]
                )

                analysis.set_pvalue(pval=pval)
            else:
                logging.debug(f"No genes found for {analysis.organism}:{analysis.pathway_id}. p value calculation skipped.")

        self.summary = result
        return result


    # def export(self):

    #     # Check if summary exists
    #     if not self.summary:
    #         raise ValueError("need to 'run_summary' first")

    #     result = list()
    #     result.append("\t".join(KEGGPathwayAnalysisResult.get_header()))

    #     # TODO export as file option
    #     # TODO use csv dict writer

    #     for item in self.summary:
    #         result.append(item.line_summary())
    #     return "\n".join(result)


    def to_dataframe(self):
        """
        Return analysis result as pandas DataFrame
        """
        try:
            import pandas

            # Check if summary exists
            self._check_analysis_result_exist()
            summary_list = []
            for result in self.summary:
                summary_list.append(result.json_summary())
            return pandas.DataFrame(summary_list)
        except ImportError:
            # print("Package 'pandas' is not installed. To use this function please 'pip install pandas'")
            logging.error("Package 'pandas' is not installed. To use this function please 'pip install pandas'")
            return None


    def fexport(self, file: Union[str, TextIOWrapper], delimiter="\t", overwrite=False):

        # Check if summary exists
        self._check_analysis_result_exist()

        if isinstance(file, str):
            # file is str (Name of file)
            if os.path.isfile(file) and not overwrite:
                raise RuntimeError(f"File {file} does already exist. To solve please set overwrite=True.")

            csv_file = open(file, mode="w", encoding="utf-8")
        elif isinstance(file, TextIOWrapper):
            # file is TextIOWrapper (File object stream)
            csv_file = file


        # Delimiter of gene names
        child_delimiter = ","

        # Change child delimiter to aviod csv conficts
        if delimiter == child_delimiter:
            child_delimiter = ";"

        headers: List[str] = KEGGPathwayAnalysisResult.get_header()
        writer = csv.DictWriter(csv_file, fieldnames=headers)

        for item in self.summary:
            # Write lines
            writer.writerow(item.json_summary(gene_delimiter=child_delimiter))

        csv_file.close()


if __name__ == "__main__":
    pass




""" Resolve requests to KEGG data Api """

import logging
from .utils import parse_tsv, request
from .storage import KEGGDataStorage
from .models import KEGGPathway


class KEGGPathwayResolver:
    """
    KEGGPathwayResolver
    Request interface for KEGG API endpoint
    """
    def __init__(self, org: str):
        """
        Need <org> 3 letter code for organism
        :param org: str
        """

        if type(org) != str:
            logging.error(f"Expect type str. Got type {type(org).__name__}")
            raise TypeError("Expect type str for organism.")

        self.organism = org


    @staticmethod
    def request(url: str):
        """
        GET request to given Url
        :param url: str
        :return: byte
        """

        return request(url=url)


    def get_pathway_list(self):
        """
        Request list of pathways linked to organism. {<pathway-id>: <name>}
        :return: dict
        """

        # path:mmu00010	Glycolysis / Gluconeogenesis - Mus musculus (mouse)
        # path:<org><code>\t<name> - <org>

        # Request list of pathways from API

        store = KEGGDataStorage()
        if store.pathway_list_exist(org=self.organism):
            # return pathway list dump
            logging.debug(f"Found pathway list for organism {self.organism} in cache.")
            data = store.load_dump(filename="pathway_{ORG}.dump".format(ORG=self.organism))
            return data

        else:
            # request pathway list
            logging.debug(f"Not found pathway list for organism {self.organism} in cache. Requesting from API")
            data = parse_tsv(KEGGPathwayResolver.request(url="http://rest.kegg.jp/list/pathway/{ORG}".format(ORG=self.organism)))
            pathways = {}
            for line in data:
                if len(line) == 2 and line[0] != "":
                    pathways[line[0].split(":")[1].strip(self.organism)] = line[1].split(" - ")[0]

            logging.debug("Loading list of {N} pathways for organism {ORG}".format(N=len(pathways.keys()), ORG=self.organism))

            # save dump
            store.save_dump(filename="pathway_{ORG}.dump".format(ORG=self.organism), data=pathways)

            # return pathway list
            return pathways


    @staticmethod
    def build_url(org: str, code: str):
        """
        Build path to KGML File at KEGG API endpint
        :param org: str
        :param code: str
        :return: str
        """
        return "http://rest.kegg.jp/get/{ORG}{CODE}/kgml".format(ORG=org, CODE=code)

    def get_pathway(self, code: str):
        """
        Request pathway by code
        :param code: str
        :return: KEGGPathway
        """
        store = KEGGDataStorage()
        if store.pathway_file_exist(org=self.organism, code=code):
            # load from file
            data = store.load(filename="{ORG}_path{CODE}.kgml".format(ORG=self.organism, CODE=code))
            logging.debug("Load pathway path:{ORG}{CODE} from file".format(ORG=self.organism, CODE=code))
        else:
            # request pathway and store
            data = KEGGPathwayResolver.request(KEGGPathwayResolver.build_url(org=self.organism, code=code))
            store.save(filename="{ORG}_path{CODE}.kgml".format(ORG=self.organism, CODE=code), data=data)
            logging.debug("Download pathway path:{ORG}{CODE} from rest.kegg.jp".format(ORG=self.organism, CODE=code))
        return KEGGPathway.parse(data)

    def link_pathways(self, geneid: str):
        """
        Return all pathways linked to gene-id
        :param geneid: str
        :return: list
        """
        data = parse_tsv(KEGGPathwayResolver.request("http://rest.kegg.jp/link/pathway/{ORG}:{ID}".format(ORG=self.organism, ID=geneid)))
        result = []
        for item in data:
            if len(item) == 2 and item[0] != "":
                result.append(item[1])
        return result

    def download_pathways(self, pathways: list):
        """
        Download all pathways from list of pathway id's.
        :param pathways:
        :return: NoneType
        """
        downloads = 0
        for code in pathways:
            if not KEGGDataStorage.pathway_file_exist(org=self.organism, code=code):
                url = KEGGPathwayResolver.build_url(org=self.organism, code=code)
                logging.debug("Requesting path:{ORG}{CODE} {URL}...".format(URL=url, ORG=self.organism, CODE=code))
                KEGGDataStorage.save(filename="{ORG}_path{CODE}.kgml".format(ORG=self.organism, CODE=code),
                                     data=KEGGPathwayResolver.request(url))
                downloads += 1
        logging.debug("Download {N} pathway KGML files from KEGG".format(N=downloads))
        return None

    @staticmethod
    def get_components():
        filename = "compound.dump"
        if not KEGGDataStorage.exist(filename=filename):
            url = "http://rest.kegg.jp/list/compound/"
            logging.debug("Requesting components {URL}...".format(URL=url))
            result = {}
            for items in parse_tsv(KEGGPathwayResolver.request(url=url)):
                if len(items) >= 2 and items[0] != "":
                    result[items[0].split(":")[1]] = items[1].split(";")[0]
            KEGGDataStorage.save_dump(filename=filename, data=result)
            return result
        else:
            return KEGGDataStorage.load_dump(filename=filename)


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    logging.info(KEGGPathwayResolver.get_components())




""" Global constants """

# Example list for immune system pathways
IMMUNE_SYSTEM_PATHWAYS = {
    "04640": "Hematopoietic cell lineage",
    "04610": "Complement and coagulation cascades",
    "04611": "Platelet activation",
    "04620": "Toll-like receptor signaling pathway",
    "04621": "NOD-like receptor signaling pathway",
    "04622": "RIG-I-like receptor signaling pathway",
    "04623": "Cytosolic DNA-sensing pathway",
    "04625": "C-type lectin receptor signaling pathway",
    "04650": "Natural killer cell mediated cytotoxicity",
    "04612": "Antigen processing and presentation",
    "04660": "T cell receptor signaling pathway",
    "04658": "Th1 and Th2 cell differentiation",
    "04659": "Th17 cell differentiation",
    "04657": "IL-17 signaling pathway",
    "04662": "B cell receptor signaling pathway",
    "04664": "Fc epsilon RI signaling pathway",
    "04666": "Fc gamma R-mediated phagocytosis",
    "04670": "Leukocyte transendothelial migration",
    "04672": "Intestinal immune network for IgA production",
    "04062": "Chemokine signaling pathway"
}


# Constant element types in the KGML format

RELATION_SUBTYPES = {
    "activation": "-->",
    "inhibition": "--|",
    "expression": "-->",
    "repression": "--|",
    "indirect effect": "..>"
}


RELATION_TYPES = {
    "ECrel": "enzyme-enzyme relation",
    "PPrel": "protein-protein interaction",
    "GErel": "gene expression interaction",
    "PCrel": "protein-compound interaction",
    "maplink": "link to another map"
}

GRAPHICS_TYPE = [
    "rectangle",
    "circle",
    "roundrectangle",
    "line"
]

ENTRY_TYPE = [
    "ortholog",
    "enzyme",
    "reaction",
    "gene",
    "group",
    "compound",
    "map",
    "brite",
    "other"
]


# KEGG cache folder name

import os

if os.environ.get("KEGG_DATA", None) != None:
    # if KEGG_DATA is set, overwrite caching dir
    KEGG_DATA = os.environ["KEGG_DATA"]
else:
    # fallback to default caching dir
    KEGG_DATA = os.path.join(os.path.dirname(__file__), ".cache")



""" Caching """

import logging
import os
import re
import pickle

from .utils import parse_tsv, request as request_url, getcwd
from .const import KEGG_DATA


class KEGGDataStorage:
    """
    Storage Handler
    Dirname of storage saved in variable KEGG_DATA
    """
    @staticmethod
    def _check_env():
        # Check if folder exists
        cache_dir = KEGG_DATA
        if not os.path.isdir(cache_dir):
            logging.debug(f"Cache directory does not exist. Creating directory {cache_dir}")
            os.mkdir(cache_dir)

    @staticmethod
    def build_path(filename: str):
        KEGGDataStorage._check_env()
        return os.path.join(KEGG_DATA, filename)

    @staticmethod
    def exist(filename: str):
        KEGGDataStorage._check_env()
        return os.path.isfile(os.path.join(KEGG_DATA, filename))


    # TODO: remove from storage  --> http request must be performed in resolver
    @staticmethod
    def get_organism_list():
        """
        Get organism codes from file or KEGG API. {<org>: <org-name>}
        :return: dict
        """
        path = KEGGDataStorage.build_path("organism.dump")
        organism_list = {}

        if not os.path.isfile(path):
            # request organism list
            result = parse_tsv(request_url(url="http://rest.kegg.jp/list/organism"))
            for item in result:
                if len(item) == 4 and item[0] != "":
                    organism_list[item[1]] = item[2]
            pickle.dump(organism_list, open(path, "wb"))
            logging.debug("Request organism list and dump to {PATH}".format(PATH=path))
        else:
            organism_list = pickle.load(open(path, "rb"))
            logging.debug("Load organism list from {PATH}".format(PATH=path))
        return organism_list

    @staticmethod
    def get_organism_name(org_code: str):
        """
        Get full name of organism by 3 letter code
        :param org_code: str
        :return: str
        """
        return KEGGDataStorage.get_organism_list().get(org_code, None)

    @staticmethod
    def check_organism(org: str):
        """
        Check if 3 letter organism code exist
        :param org: str
        :return: bool
        """
        organism_list = KEGGDataStorage.get_organism_list()
        return org in organism_list.keys()

    @staticmethod
    def list_existing_pathways():
        """
        List all KEGG pathway files saved in storage directory. [<filename>]
        :return: list
        """
        # <org>_path<code>.kgml
        found = []
        pattern = r"[a-z]{3}_path[0-9]{5}\.kgml"
        for item in os.listdir(KEGG_DATA):
            if re.match(pattern, item, re.IGNORECASE):
                found.append(item)
        logging.debug("Found {N} pathway files".format(N=len(found)))
        return found

    @staticmethod
    def pathway_file_to_id(pathway: str):
        """
        Convert pathway filename to pathway kegg-id
        :param pathway: str
        :return: str
        """
        pattern = r"[a-z]{3}_path([0-9]{5})\.kgml"
        return re.findall(pattern, pathway, re.IGNORECASE)[0]

    @staticmethod
    def pathway_to_id(pathway: str):
        """
        Extract id from pathway kegg-id
        :param pathway: str
        :return: str
        """
        pattern = r"path:[a-z]{3}([0-9]{5})"
        return re.findall(pattern, pathway, re.IGNORECASE)[0]

    @staticmethod
    def pathway_file_exist(org: str, code: str):
        """
        Check if pathway file exist in local storage
        :param org: str
        :param code: str
        :return: bool
        """
        return os.path.isfile(os.path.join(KEGG_DATA, "{ORG}_path{CODE}.kgml".format(ORG=org, CODE=code)))

    @staticmethod
    def pathway_list_exist(org: str):
        """
        Check if list of pathways exists in local stoarage
        :param org: str
        :return bool
        """
        return os.path.isfile(os.path.join(KEGG_DATA, "pathway_{ORG}.dump".format(ORG=org)))

    @staticmethod
    def load_pathway(org: str, code: str):
        """
        Load pathway string from local storage.
        :param org: str
        :param code: str
        :return: str
        """
        if KEGGDataStorage.pathway_file_exist(org=org, code=code):
            return KEGGDataStorage.load("{ORG}_path{CODE}.kgml".format(ORG=org, CODE=code))
        else:
            raise FileNotFoundError("Pathway path:{ORG}{CODE} not saved in local storage".format(ORG=org, CODE=code))

    @staticmethod
    def save(filename: str, data: str):
        """
        Save string as file in local storage. Return path.
        :param filename: str
        :param data: str
        :return: str
        """
        KEGGDataStorage._check_env()
        path = os.path.join(KEGG_DATA, filename)
        with open(path, "w") as f:
            f.write(data)
            f.close()
        return path

    @staticmethod
    def save_dump(filename: str, data):
        """
        Save binary dump as file in local storage. return path.
        :param filename: str
        :param data: str
        :return: str
        """
        KEGGDataStorage._check_env()
        path = os.path.join(KEGG_DATA, filename)
        pickle.dump(data, open(path, "wb"))
        return path

    @staticmethod
    def load(filename: str):
        """
        Load string from file.
        :param filename: str
        :return: str
        """
        KEGGDataStorage._check_env()
        path = os.path.join(KEGG_DATA, filename)
        if not os.path.isfile(path):
            raise FileNotFoundError("File at path {PATH} does not exist".format(PATH=path))
        with open(path, "r") as f:
            data = f.read()
            f.close()
        return data

    @staticmethod
    def load_dump(filename: str):
        """
        Load binary dump from file
        :param filename: str
        :return: object
        """
        KEGGDataStorage._check_env()
        path = os.path.join(KEGG_DATA, filename)
        if not os.path.isfile(path):
            raise FileNotFoundError("File at path {PATH} does not exist".format(PATH=path))
        return pickle.load(open(path, "rb"))


if __name__ == "__main__":

    logging.basicConfig(level=logging.DEBUG)

    # get all stored pathways files from folder
    pathway_files = KEGGDataStorage.list_existing_pathways()
    logging.debug("Found {N} kgml files in folder".format(N=len(pathway_files)))

    # convert pathway file string to pathway ids
    pathway_ids = [KEGGDataStorage.pathway_file_to_id(p) for p in pathway_files]
    logging.debug(len(pathway_ids), "total pathway ids found.")

    # load organism list
    logging.debug(len(KEGGDataStorage.get_organism_list().keys()), "organisms found.")

    # check if pathway file exists
    logging.debug(KEGGDataStorage.pathway_file_exist(org="mmu", code="00010"))

    # check if organism 3 letter code exists
    logging.debug(KEGGDataStorage.check_organism("mmu"))

    # get name of organism from code
    logging.debug(KEGGDataStorage.get_organism_name(org_code="mmu"))




""" KEGG pathway models to parse object relational """

import logging
from xml.etree import ElementTree as ElementTree
from xml.etree.ElementTree import Element, SubElement
from typing import Type, Union
import pydot


class Relation:
    """

    """
    def __init__(self):
        # REQUIRED
        self.entry1 = None
        self.entry2 = None
        self.type = None

        self.subtypes = {}

    @staticmethod
    def parse(item: Element):
        """
        Parse xml ElementTree into KEGG Relation
        :param item: ElementTree
        :return: Relation
        """
        relation = Relation()
        relation.entry1 = item.attrib["entry1"]
        relation.entry2 = item.attrib["entry2"]
        relation.type = item.attrib["type"]

        for child in item:
            if child.tag == "subtype":
                relation.subtypes[child.attrib["name"]] = child.attrib["value"]

        return relation

    def __str__(self):
        return "<Relation {ENTRY1}->{ENTRY2} type='{TYPE}'".format(ENTRY1=self.entry1, ENTRY2=self.entry2,
                                                                   TYPE=self.type)


class Component:
    """
    The component element is a subelement of the entry element, and is used when the entry element is a complex node;
    namely, when the type attribute value of the entry element is "group". The nodes that constitute the complex are
    specified by recurrent calls. For example, when the complex is composed of two nodes, two component elements are
    specified. The attribute of this element is as follows.
    """
    def __init__(self):
        self.id = ""

    @staticmethod
    def parse(item: Element):
        """
        Parsing ElementTree into Component
        :param item: ElementTree
        :return: Component
        """
        c = Component()
        c.id = item.attrib["id"]
        return c


class Graphics:
    """
    Graphics information for rendering
    """
    def __init__(self):
        self.attrib = {"x": 0,
                       "y": 0,
                       "width": 0,
                       "height": 0,
                       "name": "",
                       "type": "",
                       "fgcolor": "",
                       "bgcolor": ""}

    def __str__(self):
        return "<Graphics {STR}".format(STR=" ".join(["%s='%s'" % (key, val) for key, val in self.attrib.items()]))

    def __getattr__(self, item):
        return self.attrib[item]

    @staticmethod
    def parse(item: Element):
        """
        Parse xml ElementTree into KEGG Graphics
        :param item: ElementTree
        :return: Graphics
        """
        g = Graphics()
        for key, val in item.attrib.items():
            if key in g.attrib:
                g.attrib[key] = val
        return g


class Entry:
    """
    KEGG Entry
    """
    def __init__(self):
        # REQUIRED
        self.id = ""
        self.name = ""
        self.type = ""

        # IMPLIED
        self.link = ""
        self.reaction = ""
        self.graphics = None
        self.components = []

    def get_gene_id(self):
        """
        Parse variable 'name' into KEGG ID
        :return: int
        """
        return int(self.name.split(":")[1])

    def get_id(self):
        """
        Parse variable 'name' into KEGG ID
        :return: int
        """
        # TODO : remove
        return int(self.name.split(":")[1])

    @staticmethod
    def parse(item: Element):
        """
        Parsing xml ElementTree into KEGG Entry
        :param item: ElementTree
        :return: Entry
        """
        entry = Entry()
        entry.id = item.attrib["id"]
        entry.name = item.attrib["name"].split(" ")[0]
        entry.type = item.attrib["type"]

        entry.link = item.attrib.get("link", "")
        entry.reaction = item.attrib.get("reaction", "")

        for child in item:
            if child.tag == "graphics":
                entry.graphics = Graphics.parse(child)
            elif child.tag == "component":
                entry.components.append(Component.parse(child))

        return entry

    def __str__(self):
        return "<Entry id={ID} name='{NAME}' type='{TYPE}'>".format(ID=self.id, NAME=self.name, TYPE=self.type)


class KEGGPathway:
    """
    KEGG Pathway
    """
    def __init__(self):
        # REQUIRED
        self.name = ""
        self.org = ""
        self.number = ""

        # IMPLIED
        self.title = ""
        self.image = ""
        self.link = ""

        # Children
        self.relations = []
        self.entries = []
        self.reactions = []

    def get_entry_by_id(self, entry_id: Union[str, int]):
        for item in self.entries:
            if int(item.id) == int(entry_id):
                return item
        return None

    def matches(self, gene_id_list: list):
        """
        Return percent value for matching genes. [int(<gene_id>), ...]
        :param gene_id_list: list
        :return: float
        """
        # run over entry:gene_id -> return count / len(entry)

        count = 0
        for entry in self.entries:
            if entry in gene_id_list:
                count += 1
        return count / len(self.entries)

    def get_genes(self):
        """
        List all genes from pathway {<gene_id>: <gene_name>}
        :return: dict
        """
        result = {}
        for entry in self.entries:
            if entry.type == "gene":
                result[entry.get_id()] = entry.graphics.name
        logging.debug("Get {N} unique genes from pathway".format(N=len(result.keys())))
        return result

    def __str__(self):
        return "<KEGGPathway path:{ORG}{CODE} title='{TITLE}'>".format(ORG=self.org, CODE=self.number, TITLE=self.title)

    def summarize(self):
        """
        Verbose all components of pathway
        :return: None
        """
        print(self.__str__())
        for r in self.relations:
            print(r)
        for e in self.entries:
            print(e)

    @staticmethod
    def parse(data: str):
        """
        Parsing xml String in KEGG Pathway
        :param data: str
        :return: KEGGPathway
        """
        pathw = KEGGPathway()

        root = ElementTree.fromstring(data)

        pathw.name = root.attrib["name"]
        pathw.org = root.attrib["org"]
        pathw.number = root.attrib["number"]

        pathw.title = root.attrib.get("title", "")
        pathw.image = root.attrib.get("image", "")
        pathw.link = root.attrib.get("link", "")

        for child in root:
            if child.tag == "entry":
                pathw.entries.append(Entry.parse(child))
            elif child.tag == "relation":
                pathw.relations.append(Relation.parse(child))
            elif child.tag == "reaction":
                # TODO: implement parsing
                pass
            else:
                logging.debug(child.tag)

        return pathw


if __name__ == "__main__":
    pass




""" Basic utils for HTTP requests, parsing and rendering """


import logging
from tqdm import tqdm
import requests
import math
import os
from datetime import datetime
import csv
from io import StringIO
from typing import Optional, Union



def getcwd():
    # return os.path.split(__file__)[0]
    return os.path.dirname(__file__)


def get_timestamp():
    return datetime.now().timestamp()


# Code using from
# https://stackoverflow.com/questions/312443/how-do-you-split-a-list-into-evenly-sized-chunks
# Thanks to Ned Batchelder (https://stackoverflow.com/users/14343/ned-batchelder)
def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]


class Downloader:
    def __init__(self, url: str):
        self.url = url
        self.output = ""

    def set_output(self, filename: str):
        self.output = filename

    def run(self):
        if not self.output:
            raise ValueError("output not set!")

        # Streaming, so we can iterate over the response.
        r = requests.get(self.url, stream=True)

        # Total size in bytes.
        total_size = int(r.headers.get('content-length', 0))
        block_size = 1024
        wrote = 0

        with open(self.output, 'wb') as f:
            for data in tqdm(r.iter_content(block_size),
                             total=math.ceil(total_size // block_size),
                             unit='KB',
                             unit_scale=True):

                wrote = wrote + len(data)
                f.write(data)
        if total_size != 0 and wrote != total_size:
            logging.error("ERROR, something went wrong")




def parse_tsv(data: str):
    """
    Parse .tsv file from string
    :param data: str
    :return: list
    """
    if type(data) == str:
        fstream = StringIO(data)
    else:
        raise TypeError("Type of 'data' must be str or bytes.")

    return list(csv.reader(fstream, delimiter="\t"))



def request(url: str, encoding="utf-8"):
    logging.info(f"Requesting Url {url}")
    response = requests.get(url=url)
    logging.info(f"Request finished with status code {response.status_code}")
    response.raise_for_status()
    content = response.content

    if type(content) == bytes:
        return content.decode(encoding)
    else:
        return content


class ColorGradient:
    """
    Create color gradient
    """

    def __init__(self, start: tuple, stop: tuple, steps: int):
        self.start = start
        self.stop = stop
        self.steps = steps


    @staticmethod
    def to_css(color: tuple):
        return "rgb({R},{G},{B})".format(R=color[0], G=color[1], B=color[2])


    def get_list(self):
        return [ColorGradient.list_to_hex(code) for code in ColorGradient._gradient(self.stop, self.start, self.steps)]


    @staticmethod
    def _intermediate(a, b, ratio):

        def _array_multiply(array, c):
            return [element * c for element in array]

        a_component = _array_multiply(a, ratio)
        b_component = _array_multiply(b, 1 - ratio)
        return list(map(sum, zip(a_component, b_component)))


    @staticmethod
    def _gradient(a, b, steps):
        steps -= 1
        steps = [n / float(steps) for n in range(steps)]
        result = [ColorGradient._intermediate(a, b, step) for step in steps]
        result.append(a)
        return result


    @staticmethod
    def list_to_hex(code: list):
        return "#" + "".join(["{:02x}".format(int(num)) for num in code])


    def render_graphviz(self):
        result = self.get_list()
        string = ["digraph G {"]
        for i in range(0, 20):
            # result[i]
            string.append("\tnode[color = \"{COLOR}\" style = filled] {N};".format(COLOR=result[i], N=i))
        string.append("\t" + " -> ".join([str(l) for l in range(0, 20)]))
        string.append("}")
        return "\n".join(string)


if __name__ == "__main__":

    logging.basicConfig(level=logging.DEBUG)

    c = ColorGradient(start=(0, 179, 0), stop=(187, 0, 0), steps=20)
    print(c.get_list())
    print(c.render_graphviz())

    request("http://example.com/")





import os
import setuptools
from distutils.core import setup

def read(fname: str):
    return open(os.path.join(os.path.dirname(__file__), fname), mode="r", encoding="utf-8").read()

setup(
    name = "keggtools",
    version = "0.4.0",
    license="MIT",
    author = "harryhaller001",
    author_email = "harryhaller001@gmail.com",
    description = ("Enrichment analysis and visualisation toolkit for KEGG pathways"),
    packages=setuptools.find_packages(),
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    url="https://github.com/harryhaller001/keggtools",
    install_requires=[
        "tqdm",
        "requests",
        "pydot",
        "scipy"
    ],
    # download_url="https://github.com/harryhaller001/keggtools" # TODO: add release
)

